package com.example.aida

import android.content.Context
import android.content.Intent
import android.os.Bundle
import android.speech.tts.TextToSpeech
import androidx.activity.ComponentActivity
import androidx.activity.compose.setContent
import androidx.compose.foundation.layout.Spacer
import androidx.compose.foundation.layout.padding
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.automirrored.filled.ExitToApp
import androidx.compose.material.icons.automirrored.filled.MenuBook
import androidx.compose.material.icons.filled.Home
import androidx.compose.material.icons.filled.Settings
import androidx.compose.material3.DrawerValue
import androidx.compose.material3.HorizontalDivider
import androidx.compose.material3.Icon
import androidx.compose.material3.MaterialTheme
import androidx.compose.material3.ModalDrawerSheet
import androidx.compose.material3.ModalNavigationDrawer
import androidx.compose.material3.NavigationDrawerItem
import androidx.compose.material3.Surface
import androidx.compose.material3.Text
import androidx.compose.material3.rememberDrawerState
import androidx.compose.runtime.Composable
import androidx.compose.runtime.getValue
import androidx.compose.runtime.mutableIntStateOf
import androidx.compose.runtime.mutableStateOf
import androidx.compose.runtime.remember
import androidx.compose.runtime.rememberCoroutineScope
import androidx.compose.runtime.setValue
import androidx.compose.ui.Modifier
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.platform.LocalConfiguration
import androidx.compose.ui.platform.LocalContext
import androidx.compose.ui.unit.dp
import androidx.datastore.core.DataStore
import androidx.datastore.preferences.core.Preferences
import androidx.datastore.preferences.preferencesDataStore
import androidx.lifecycle.viewmodel.compose.viewModel
import com.example.aida.pages.CameraPage
import com.example.aida.pages.ConfigurationPage
import com.example.aida.pages.UserGuidePage
import com.example.aida.pages.SequenceTab
import com.example.aida.ui.composables.TopBar
import com.example.aida.ui.theme.AIDATheme
import com.example.aida.viewmodels.MainViewModel
import com.example.aida.viewmodels.MainViewModelFactory
import kotlinx.coroutines.launch
import androidx.compose.ui.tooling.preview.Preview
import androidx.navigation.NavHostController
import androidx.navigation.compose.NavHost
import androidx.navigation.compose.composable
import androidx.navigation.compose.rememberNavController

import com.example.aida.viewmodels.SplashScreen
import java.util.Locale
import androidx.compose.foundation.layout.Row
import androidx.compose.material3.IconButton
import androidx.compose.material.icons.filled.Close
import androidx.compose.ui.Alignment

private val Context.dataStore: DataStore<Preferences> by preferencesDataStore(
    name = "settings"
)

enum class State {
    HOME,
    SETTINGS,
    USER_GUIDE,
    INPUT
}

class MainActivity : ComponentActivity(), TextToSpeech.OnInitListener {

    private lateinit var viewModel: MainViewModel
    var textToSpeech: TextToSpeech? = null  // Add a reference to TextToSpeech

    override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)
        // Initialize TTS
        textToSpeech = TextToSpeech(this, this)

        setContent {
            AIDATheme {
                MyAppNavHost()
            }
        }
    }

    override fun onDestroy() {
        textToSpeech?.stop()
        textToSpeech?.shutdown()
        viewModel.closeConnections()
        super.onDestroy()
    }

    // TTS initialization callback
    override fun onInit(status: Int) {
        if (status == TextToSpeech.SUCCESS) {
            textToSpeech?.language = Locale.getDefault() // Set default language
        }
    }

    override fun onActivityResult(requestCode: Int, resultCode: Int, data: Intent?) {
        super.onActivityResult(requestCode, resultCode, data)
    }

    @Composable
    fun AppContent(viewModel: MainViewModel) {
        var topBarTitle by remember { mutableStateOf("AIDA Remote Control") }
        var resetState by remember { mutableStateOf(false) }
        val drawerState = rememberDrawerState(initialValue = DrawerValue.Closed)
        var state by remember { mutableStateOf(State.HOME) }
        val scope = rememberCoroutineScope()

        ModalNavigationDrawer(
            drawerState = drawerState,
            gesturesEnabled = false,
            drawerContent = {
                ModalDrawerSheet {
                    Row(
                        verticalAlignment = Alignment.CenterVertically,
                        modifier = Modifier.padding(16.dp)
                    ) {
                        Text("Menu", modifier = Modifier.weight(1f))
                        IconButton(onClick = {
                            scope.launch {
                                drawerState.close()
                            }
                        }) {
                            Icon(Icons.Filled.Close, contentDescription = "Close Drawer")
                        }
                    }
                    HorizontalDivider()
                    NavigationDrawerItem(
                        icon = { Icon(Icons.Filled.Home, contentDescription = "Home") },
                        label = { Text(text = "Home") },
                        selected = false,
                        onClick = {
                            state = State.HOME
                            scope.launch {
                                drawerState.close()
                            }
                            topBarTitle = "AIDA Remote Control"
                        }
                    )
                    NavigationDrawerItem(
                        icon = {
                            Icon(Icons.Filled.Settings, contentDescription = "Settings")
                        },
                        label = { Text(text = "Connection Settings") },
                        selected = false,
                        onClick = {
                            state = State.SETTINGS
                            scope.launch {
                                drawerState.close()
                            }
                            topBarTitle = "Connection Settings"
                        }
                    )
                    NavigationDrawerItem(
                        icon = {
                            Icon(Icons.AutoMirrored.Filled.MenuBook, contentDescription = "User Guide")
                        },
                        label = { Text(text = "User Guide") },
                        selected = false,
                        onClick = {
                            state = State.USER_GUIDE
                            scope.launch {
                                drawerState.close()
                            }
                            topBarTitle = "User Guide"
                        }
                    )

                    Spacer(modifier = Modifier.weight(1f))
                    NavigationDrawerItem(
                        icon = {
                            Icon(
                                Icons.AutoMirrored.Filled.ExitToApp,
                                contentDescription = "Shutdown application"
                            )
                        },
                        label = { Text(text = "Application shutdown", color = Color.Red) },
                        selected = false,
                        onClick = { System.exit(0) }
                    )
                }
            },
        ) {
            Surface(
                color = MaterialTheme.colorScheme.background
            ) {
                val configuration = LocalConfiguration.current
                val screenHeight = configuration.screenHeightDp.dp
                val screenWidth = configuration.screenWidthDp.dp

                val barHeight = if (screenHeight / 8 < 50.dp) screenHeight / 6 else 50.dp

                TopBar(
                    onMenuClicked = {
                        scope.launch {
                            drawerState.apply {
                                if (isClosed) open() else close()
                            }
                        }
                    },
                    onCameraClicked = { viewModel.toggleCameraFeed() },
                    onGestureClicked = { viewModel.toggleGestureFeed() },
                    barHeight = barHeight,
                    topBarTitle = topBarTitle,
                    viewModel = viewModel,
                    resetState = resetState,
                    onResetState = { resetState = false },
                    onToggleClicked = { toggled ->
                        state = if (toggled) State.INPUT else State.HOME
                    },
                    toggleState = (state == State.INPUT)
                )

                when (state) {
                    State.HOME -> CameraPage(
                        screenHeight = screenHeight,
                        barHeight = barHeight,
                        screenWidth = screenWidth,
                        viewModel = viewModel,
                    )

                    State.SETTINGS -> ConfigurationPage(
                        barHeight = barHeight,
                        viewModel = viewModel,
                        onButtonPress = {
                            state = State.HOME
                            topBarTitle = "Configuration Page"
                            resetState = true
                        }
                    )

                    State.USER_GUIDE -> UserGuidePage(
                        barHeight = barHeight,
                        activity = this
                    )

                    State.INPUT -> SequenceTab(
                        screenHeight = screenHeight,
                        barHeight = barHeight,
                        viewModel = viewModel,
                        // Pass a lambda to handle TTS playback
                        onPlayTTS = { text ->
                            textToSpeech?.speak(text, TextToSpeech.QUEUE_FLUSH, null, "utteranceId")
                        }
                    )
                }
            }
        }
    }

    @Composable
    fun MyAppNavHost(navController: NavHostController = rememberNavController()) {
        NavHost(navController = navController, startDestination = "splash") {
            composable("splash") {
                SplashScreen(navigateToMain = {
                    navController.navigate("main") {
                        popUpTo("splash") { inclusive = true }
                    }
                })
            }
            composable("main") {
                MainScreen()
            }
        }
    }

    @Composable
    fun MainScreen() {
        val context = LocalContext.current
        val dataStore = context.dataStore
        val viewModel: MainViewModel = viewModel(factory = MainViewModelFactory(dataStore))

        AppContent(viewModel)
    }

    @Preview(showBackground = true)
    @Composable
    fun DefaultPreview() {
        AIDATheme {
            MyAppNavHost()
        }
    }
}
