package com.example.aida.pages

import android.util.Log
import androidx.annotation.OptIn
import androidx.media3.common.util.UnstableApi
import androidx.compose.foundation.ScrollState
import androidx.compose.foundation.background
import androidx.compose.foundation.gestures.detectDragGestures
import androidx.compose.foundation.gestures.detectTapGestures
import androidx.compose.foundation.layout.Arrangement
import androidx.compose.foundation.layout.Box
import androidx.compose.foundation.layout.Column
import androidx.compose.foundation.layout.Row
import androidx.compose.foundation.layout.fillMaxHeight
import androidx.compose.foundation.layout.fillMaxSize
import androidx.compose.foundation.layout.height
import androidx.compose.foundation.layout.offset
import androidx.compose.foundation.layout.padding
import androidx.compose.foundation.layout.size
import androidx.compose.foundation.layout.width
import androidx.compose.foundation.lazy.grid.GridCells
import androidx.compose.foundation.lazy.grid.LazyVerticalGrid
import androidx.compose.foundation.lazy.grid.items
import androidx.compose.foundation.rememberScrollState
import androidx.compose.foundation.shape.RoundedCornerShape
import androidx.compose.material3.Icon
import androidx.compose.material3.IconButton
import androidx.compose.material3.Text
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.outlined.*
import androidx.compose.runtime.Composable
import androidx.compose.runtime.derivedStateOf
import androidx.compose.runtime.getValue
import androidx.compose.runtime.mutableFloatStateOf
import androidx.compose.runtime.mutableStateListOf
import androidx.compose.runtime.mutableStateOf
import androidx.compose.runtime.remember
import androidx.compose.runtime.rememberCoroutineScope
import androidx.compose.runtime.setValue
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.draw.clip
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.input.pointer.pointerInput
import androidx.compose.ui.layout.onGloballyPositioned
import androidx.compose.ui.layout.positionInRoot
import androidx.compose.ui.platform.LocalDensity
import androidx.compose.ui.text.font.FontWeight
import androidx.compose.ui.unit.Dp
import androidx.compose.ui.unit.IntOffset
import androidx.compose.ui.unit.dp
import androidx.compose.ui.window.Popup
import androidx.compose.ui.zIndex
import androidx.compose.ui.draw.shadow
import androidx.compose.ui.text.style.TextAlign
import androidx.compose.ui.graphics.vector.ImageVector
import androidx.compose.animation.core.FastOutSlowInEasing
import androidx.compose.animation.core.tween
import androidx.compose.animation.core.LinearEasing
import androidx.compose.animation.animateColorAsState
import androidx.compose.runtime.mutableDoubleStateOf
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.launch
import kotlinx.coroutines.delay
import kotlin.math.roundToInt

import com.example.aida.viewmodels.MainViewModel
import com.example.aida.ui.composables.sequenceTabComposables.BaseAction
import com.example.aida.ui.composables.sequenceTabComposables.SequenceBar
import com.example.aida.ui.composables.sequenceTabComposables.SpecialAction
import com.example.aida.ui.composables.sequenceTabComposables.MoveAction
import com.example.aida.ui.composables.sequenceTabComposables.ActionLibraryData
import com.example.aida.ui.composables.sequenceTabComposables.moveActionColor
import com.example.aida.ui.composables.sequenceTabComposables.specialActionColor

/**
 * Displays the "SequenceTab" screen which includes:
 * - A sequence bar where actions can be arranged and manipulated.
 * - An action library to drag and drop actions into the sequence bar.
 * - User control buttons (play, pause, stop) to interact with the sequence.
 *
 * @param screenHeight The total height available for this tab.
 * @param barHeight The height allocated to the top bar, so the sequence tab content can be placed below it.
 */
@OptIn(UnstableApi::class)
@Composable
fun SequenceTab(
    screenHeight: Dp,
    barHeight: Dp,
    viewModel: MainViewModel,
    onPlayTTS: (String) -> Unit  // New parameter
) {
    // Mutable lists to store actions in the sequence and actions to be executed (to ROS)
    val actionsInSequenceBar = remember { mutableStateListOf<BaseAction>() }

    // Colors representing various states
    val defaultColor = Color(0xFFFFFFFF)
    val draggingColor = Color(0xFF808080)
    val playingColor = Color(0xCCADADAD)

    // The sequenceBarColor changes when dragging or playing; animated for smooth transitions
    var sequenceBarColor by remember { mutableStateOf(Color(0xFFE3E2DF)) }
    val animatedColor by animateColorAsState(targetValue = sequenceBarColor)

    // Convert Dp dimensions to pixels
    val density = LocalDensity.current
    val screenHeightPx = with(density) { screenHeight.toPx() }

    // Load action library data as a list
    val libraryActions = ActionLibraryData.toList()

    // Scroll state to handle horizontal scrolling of the sequence bar
    val scrollState = rememberScrollState()
    val coroutineScope = rememberCoroutineScope()

    // Controls whether scrolling and interaction with the sequence bar is enabled
    var scrollEnabled by remember { mutableStateOf(true) }


    // Main container for the sequence page
    Column(
        modifier = Modifier
            .padding(top = barHeight)
            .height(screenHeight - barHeight)
            .fillMaxSize()
    ) {


        // SEQUENCE BAR
        Box(
            modifier = Modifier
                .weight(1f)
                .fillMaxSize(),
            contentAlignment = Alignment.Center
        ) {
            // The sequence bar displays and manages actions currently in the sequence
            SequenceBar(
                actionsInSequenceBar = actionsInSequenceBar,
                sequenceBarColor = animatedColor,
                scrollState = scrollState,
                coroutineScope = coroutineScope,
                scrollEnabled = scrollEnabled,
                onPlayTTS = onPlayTTS
            )
        }

        // ACTION LIBRARY AND USER BUTTONS SECTION
        Row(
            modifier = Modifier
                .weight(1f)
                .background(color = Color.LightGray)
                .fillMaxSize()
        ) {
            // ACTION LIBRARY: A grid of actions that can be dragged into the sequence bar
            Box(
                modifier = Modifier
                    .weight(0.5f)
                    .fillMaxHeight()
                    .width(200.dp),
                contentAlignment = Alignment.Center
            ) {
                if(!scrollEnabled)
                {
                    // If scrolling is disabled (e.g., while playing), overlay a transparent box
                    // to block interaction with the action library.
                    Box(
                        modifier = Modifier
                            .background(Color.Transparent)
                            .fillMaxSize()
                            .zIndex(4f)
                            .pointerInput(Unit) {
                                detectTapGestures { /* no-op, just block interaction */ }
                            }
                    ) {}
                }
                // Display the action library as a grid of draggable buttons
                LazyVerticalGrid(
                    columns = GridCells.Fixed(7),
                    verticalArrangement = Arrangement.spacedBy(8.dp),
                    horizontalArrangement = Arrangement.spacedBy(8.dp),
                    modifier = Modifier
                        .padding(start = 8.dp, top = 20.dp, end = 8.dp, bottom = 8.dp)
                        .fillMaxSize()
                ) {
                    items(libraryActions) { action ->
                        // Only display actions that are not "PAD" (they are used as padding)
                        if (action.name != "PAD") {
                            DraggableButton(
                                text = action.name,
                                icon = action.icon.value,
                                cardColor = action.cardColor.value,
                                color = if (action is MoveAction) moveActionColor else specialActionColor,
                                onDragStart = {
                                    // Change color to indicate dragging state
                                    sequenceBarColor = draggingColor
                                },
                                onDragEnd = { globalOffsetY ->
                                    // Reset color after drag ends
                                    sequenceBarColor = defaultColor
                                    // If dropped in the top half of the screen (where sequence bar is),
                                    // add the action to the sequence bar.
                                    if (globalOffsetY < screenHeightPx / 2) {
                                        addActionToSequenceBar(action, actionsInSequenceBar, scrollState, coroutineScope)
                                    }
                                },
                                onSingleClick = {
                                    // Single click directly adds the action to the sequence bar
                                    addActionToSequenceBar(action, actionsInSequenceBar, scrollState, coroutineScope)
                                },
                                onDoubleClick = {
                                    // Double click also adds the action to the sequence bar
                                    addActionToSequenceBar(action, actionsInSequenceBar, scrollState, coroutineScope)
                                },
                            )
                        }
                    }
                }
            }
            // USER BUTTONS: Play, stop, step, etc.
            UserButtons(
                scrollState = scrollState,
                coroutineScope = coroutineScope,
                actionsInSequenceBar = actionsInSequenceBar,
                sequenceBarColor = sequenceBarColor,
                updateSequenceBarColor = { updatedValue ->
                    sequenceBarColor = updatedValue
                },
                defaultColor = defaultColor,
                playingColor = playingColor,
                scrollEnabled = scrollEnabled,
                updateScrollEnabled = { updatedValue ->
                    scrollEnabled = updatedValue
                },
                viewModel = viewModel
            )
        }
    }
}

/**
 * A draggable button representing an action in the action library.
 *
 * @param text The display text of the action.
 * @param icon The icon representing the action.
 * @param color The background color of the button.
 * @param onDragStart Callback when dragging starts.
 * @param onDragEnd Callback when dragging ends, providing the Y offset in pixels of the final position.
 * @param onSingleClick Callback when the button is single-clicked.
 * @param onDoubleClick Callback when the button is double-clicked.
 */
@Composable
fun DraggableButton(
    text: String,
    icon: ImageVector,
    cardColor: Color,
    color: Color,
    onDragStart: (() -> Unit)? = null,
    onDragEnd: ((Float) -> Unit)? = null,
    onSingleClick: (() -> Unit)? = null,
    onDoubleClick: (() -> Unit)? = null,
) {
    val buttonWidth = 130.dp
    val buttonHeight = 150.dp
    var offsetX by remember { mutableFloatStateOf(0f) }
    var offsetY by remember { mutableFloatStateOf(0f) }
    var isDragging by remember { mutableStateOf(false) }
    var globalOffsetY by remember { mutableFloatStateOf(0f) }

    var isPressed by remember { mutableStateOf(false) } // State to track if the button is pressed

    val density = LocalDensity.current
    Box(
        modifier = Modifier
            // Positioning the box based on offsetX/offsetY (for dragging)
            .offset { IntOffset(offsetX.roundToInt(), offsetY.roundToInt()) }
            // If dragging, collapse the original size to 0 to prevent overlap; otherwise normal size
            .then(
                if (isDragging) Modifier.size(0.dp) else Modifier.size(
                    buttonWidth,
                    buttonHeight
                )
            )
            .clip(RoundedCornerShape(8.dp))
            .background(color = if (isDragging) Color.Gray else color)
            .shadow(
                elevation = 0.dp,
                shape = RoundedCornerShape(8.dp),
                spotColor = Color.Gray,
                ambientColor = Color.Black,
            )
            .onGloballyPositioned { coordinates ->
                // Track the global Y offset of this button
                globalOffsetY = coordinates.positionInRoot().y
            }
            // Detect drag gestures
            .pointerInput(Unit) {
                detectDragGestures(
                    onDragStart = {
                        isDragging = true
                        onDragStart?.invoke()
                    },
                    onDragEnd = {
                        // When drag ends, reset dragging state and call onDragEnd
                        isDragging = false
                        onDragEnd?.invoke(globalOffsetY)
                        offsetX = 0f
                        offsetY = 0f
                    },
                    onDragCancel = {
                        // If drag is canceled, just stop dragging
                        isDragging = false
                    },
                    onDrag = { change, dragAmount ->
                        change.consume()

                        // Convert dragAmount from pixels to DP using screen density
                        with(density) {
                            offsetX += dragAmount.x.toDp().value
                            offsetY += dragAmount.y.toDp().value
                        }
                    }
                )
            }
            // Detect tap and double-tap gestures
            .pointerInput(Unit) {
                detectTapGestures(
                    onPress = {
                        isPressed = true
                        tryAwaitRelease()
                        isPressed = false
                    },
                    onTap = {
                        onSingleClick?.invoke()
                    },
                    onDoubleTap = {
                        onDoubleClick?.invoke()
                    }
                )
            },
        contentAlignment = Alignment.Center
    ) {
        // Button content: Icon and Text
        Column(
            modifier = Modifier
                .fillMaxSize()
                .padding(10.dp),
            horizontalAlignment = Alignment.CenterHorizontally,
        ) {
            Box(
                modifier = Modifier
                    .size(50.dp)
                    .weight(1.5f),
                contentAlignment = Alignment.Center
            ){
                Icon(
                    imageVector = icon,
                    contentDescription = "",
                    tint = Color.White,
                    modifier = Modifier
                        .fillMaxSize()
                )
            }

            Box(
                modifier = Modifier
                    .weight(1f),
                contentAlignment = Alignment.Center
            ){
                Text(
                    text = text,
                    color = Color.White,
                    textAlign = TextAlign.Center,
                    fontWeight = FontWeight.Bold,
                )
            }
        }
        // If currently dragging, show a popup representation at the drag location
        if (isDragging){
            Popup(
                alignment = Alignment.TopStart,
                offset = IntOffset(offsetX.roundToInt(), offsetY.roundToInt())
            ){
                Box(
                    modifier = Modifier
                        .width(147.dp)
                        .height(buttonHeight)
                        .clip(RoundedCornerShape(8.dp))
                        .background(color = cardColor),
                    contentAlignment = Alignment.Center
                ) {
                // Button content: Icon and Text
                    Column(
                        modifier = Modifier
                            .fillMaxSize()
                            .padding(10.dp),
                        horizontalAlignment = Alignment.CenterHorizontally,
                    ) {
                        Box(
                            modifier = Modifier
                                .size(50.dp)
                                .weight(1.5f),
                            contentAlignment = Alignment.Center
                        ){
                            Icon(
                                imageVector = icon,
                                contentDescription = "",
                                tint = Color.White,
                                modifier = Modifier
                                    .fillMaxSize()
                            )
                        }

                        Box(
                            modifier = Modifier
                                .weight(1f),
                            contentAlignment = Alignment.Center
                        ){
                            Text(
                                text = text,
                                color = Color.White,
                                textAlign = TextAlign.Center,
                                fontWeight = FontWeight.Bold,
                            )
                        }
                    }
                }
            }
        }
    }
}

/**
 * Displays the user control buttons (Stop, Step, Play/Pause).
 *
 * Manages the playing state, looping logic, and scroll animations.
 * Also handles sending actions to a ROS (Robot Operating System) backend.
 *
 * @param scrollState The [ScrollState] controlling the sequence bar's position.
 * @param coroutineScope A [CoroutineScope] for launching animation and action tasks.
 * @param actionsInSequenceBar The current list of actions in the sequence bar.
 * @param sequenceBarColor Current background color of the sequence bar.
 * @param updateSequenceBarColor Callback to update the sequence bar color.
 * @param defaultColor The default (idle) color of the sequence bar.
 * @param playingColor The color of the sequence bar while playing actions.
 * @param actionsToRos A mutable list representing actions sent to ROS for execution.
 * @param scrollEnabled Whether scrolling the sequence bar is currently enabled.
 * @param updateScrollEnabled Callback to enable/disable scrolling.
 */
@Composable
fun UserButtons(
    scrollState: ScrollState,
    coroutineScope: CoroutineScope,
    actionsInSequenceBar: List<BaseAction>,
    sequenceBarColor: Color,
    updateSequenceBarColor: (Color) -> Unit,
    defaultColor: Color,
    playingColor: Color,
    scrollEnabled: Boolean,
    updateScrollEnabled: (Boolean) -> Unit,
    viewModel: MainViewModel
) {
    var playPause by remember { mutableStateOf(false) }
    val icon = if (playPause) Icons.Outlined.PauseCircle else Icons.Outlined.NotStarted
    var userStates by remember { mutableStateOf("STOP") } // track the user state

    // Button colors depending on the user state
    val stopButtonColor by remember {
        derivedStateOf { Color(0xFF8C241A) }
    }
    val stepButtonColor by remember {
        derivedStateOf { if (userStates == "PLAY" || userStates == "STEP") Color.LightGray else Color(0xFF6DA925) }
    }
    val playButtonColor by remember {
        derivedStateOf { if (userStates == "STOP" || userStates == "PLAY") Color(0xFFFFC300) else Color.LightGray }
    }

    val stepvalue = 258 // Represents the scroll distance between actions
    var curr = 0
    var remainder = 0
    var loopStartIndex = 0
    var loopCounter = 0
    var isLooping by remember { mutableStateOf(false) }

    // Flags representing each button state
    var isPause by remember { mutableStateOf(false) }
    var isStep by remember { mutableStateOf(false) }
    var isPlay by remember { mutableStateOf(false) }
    var isStop by remember { mutableStateOf(false) }


    Box(
        modifier = Modifier
            .fillMaxHeight()
            .width(200.dp)
            .padding(15.dp),
        contentAlignment = Alignment.Center
    ) {
        Box(
            modifier = Modifier
                .fillMaxSize()
                .clip(RoundedCornerShape(20.dp))
        ) {
            Column(
                horizontalAlignment = Alignment.CenterHorizontally,
                verticalArrangement = Arrangement.SpaceEvenly,
                modifier = Modifier
                    .fillMaxSize()
                    .background(Color.Gray)
            ) {
                // STOP BUTTON
                IconButton(
                    onClick = {
                        if((actionsInSequenceBar.isNotEmpty()))
                        {
                            if(isPlay || isStop) {
                                // If already playing or stopped, pressing stop again resets states
                                isStop = true
                                isLooping = false
                                resetAllActionDuration(actionsInSequenceBar)
                                userStates = "STOP"

                                // Stop sending actions to ROS
                                if (viewModel.isSequenceConnected) {
                                    viewModel.sendStopExecutingActions()
                                }
                            }
                            else {
                                // If not playing or stepping, pressing stop scrolls to start
                                coroutineScope.launch {
                                    updateScrollEnabled(false)
                                    updateSequenceBarColor(playingColor)
                                    scrollState.animateScrollTo(0, animationSpec = tween(durationMillis = 750, easing = FastOutSlowInEasing))
                                    updateSequenceBarColor(defaultColor)
                                    updateScrollEnabled(true)
                                }
                            }
                        }
                    },
                    modifier = Modifier.size(70.dp)
                ) {
                    Icon(
                        imageVector = Icons.Outlined.StopCircle,
                        contentDescription = "Stop",
                        tint = stopButtonColor,
                        modifier = Modifier
                            .fillMaxSize()
                    )
                }
                // STEP BUTTON
                IconButton(
                    onClick = {
                        if(!isPlay && !isStop && !isStep && (actionsInSequenceBar.isNotEmpty()))
                        {
                            isStep = true
                            isLooping = false
                            userStates = "STEP"
                            curr = (scrollState.value/stepvalue) + 1
                            remainder = scrollState.value % stepvalue

                            coroutineScope.launch {
                                updateScrollEnabled(false)
                                updateSequenceBarColor(playingColor)
                                // Execute the current action
                                if((curr - 1) < actionsInSequenceBar.size)
                                {
                                    // Execute action and countdown
                                   executeActionWithCountdown(actionsInSequenceBar[curr - 1], viewModel, coroutineScope)
                                }
                                else
                                {
                                    Log.e("SequenceTab", "Warning: Reached end of Sequence")
                                }

                                // Animate scroll to next action
                                // the delay is to simulate the robots actions
                                // delay will later depend on how quickly the robot perform its actions
                                if(scrollState.value > ((curr * stepvalue) - (stepvalue/2)))
                                {
                                    scrollState.animateScrollTo((curr) * stepvalue)
                                    delay(500)
                                    scrollState.animateScrollTo((((curr + 1) * stepvalue)), animationSpec = tween(durationMillis = 1000, easing = LinearEasing))
                                }
                                else
                                {
                                    scrollState.animateScrollTo((((curr - 1) * stepvalue)))
                                    delay(500)
                                    scrollState.animateScrollTo((((curr) * stepvalue)), animationSpec = tween(durationMillis = 1000, easing = LinearEasing))
                                }
                                isStep = false
                                userStates = "STOP"
                                updateSequenceBarColor(defaultColor)
                                updateScrollEnabled(true)
                                resetAllActionDuration(actionsInSequenceBar)
                            }
                            Log.d("ScrollState", "Current Scroll Position: $curr")
                            Log.d("ActionIndex", "Current Action: ${(scrollState.value/stepvalue) + 1}")
                        }
                    },
                    modifier = Modifier.size(70.dp)
                ) {
                    Icon(
                        imageVector = Icons.Outlined.PlayCircle,
                        contentDescription = "not started",
                        tint = if (!playPause) stepButtonColor else Color.LightGray,
                        modifier = Modifier.fillMaxSize()
                    )
                }
                // !!! WARNING BELOW THIS LINE !!!
                //
                // PLAY/PAUSE BUTTON
                //
                // FOR THE GROUP CONTINUING WORK ON THIS PROJECT
                //
                // THE PLAY / PAUSE FUNCTIONALITY IS ONLY A PSEUDO - VISUAL REPRESENTATION OF HOW THE APP IS THOUGHT TO WORK
                //
                // PLEASE REMAKE THE PLAY / PAUSE AND STEP FUNCTIONALITY ACCORDING TO YOUR NEEDS!
                //
                // our groups thoughts are that when an instruction is sent to RoS,
                // the scrolling stops until a sort of acknowledgement is received from RoS that the instruction is executed,
                // at which point the scroll can continue!
                //
                // !!!                          !!!
                IconButton(
                    onClick = {
                        if(!isStep && !isStop && !isPause && (actionsInSequenceBar.isNotEmpty()))
                        {
                            if (userStates == "PLAY")
                            {
                                // If currently playing, pressing this toggles to pause
                                isPause = true
                                isPlay = false
                                isLooping = false

                                userStates = "STOP"

                            } else {
                                // If not currently playing, start playing
                                userStates = "PLAY"
                                isPause = false
                                isPlay = true
                                playPause = !playPause
                            }

                            coroutineScope.launch {
                                updateScrollEnabled(false)
                                updateSequenceBarColor(playingColor)
                                curr = (scrollState.value/stepvalue) + 1
                                var action = actionsInSequenceBar[curr - 1]
                                remainder = scrollState.value % stepvalue
                                Log.d("ScrollState", "Current Scroll Position: ${scrollState.value}")

                                // Adjust scroll position to action boundary
                                if(scrollState.value > ((curr * stepvalue) - (stepvalue/2)))
                                {
                                    scrollState.animateScrollTo((((curr) * stepvalue)))
                                }
                                else
                                {
                                    scrollState.animateScrollTo((((curr - 1) * stepvalue)))
                                }

                                // Continuous play loop
                                while((curr - 1) < actionsInSequenceBar.size)
                                {


                                    if(isStop)
                                    {
                                        // If stop is triggered, reset to start

                                        scrollState.animateScrollTo(0, animationSpec = tween(durationMillis = 750, easing = FastOutSlowInEasing))
                                        isStop = false
                                        isPause = false
                                        break
                                    }
                                    else if (isPause)
                                    {
                                        // If pause is triggered, break out of loop

                                        isPause = false
                                        break
                                    }
                                    else
                                    {
                                        // Execute next action

                                        if((curr - 1) < actionsInSequenceBar.size)
                                        {
                                            // Sends action to ROS
                                            executeActionWithCountdown(
                                                actionsInSequenceBar[curr - 1],
                                                viewModel,
                                                coroutineScope
                                            )
                                        }
                                        else
                                        {
                                            Log.e("SequenceTab", "Warning: Reached end of Sequence")
                                        }
                                        action = actionsInSequenceBar[curr - 1]
                                        // If loop start encountered
                                        if (action.name == "LOOP START")
                                        {
                                            resetActionDurationInRange(actionsInSequenceBar,  curr-1, actionsInSequenceBar.size - 1)
                                            println("IN loop Start")
                                            if (!isLooping)
                                            {
                                                loopCounter = (action.data.value).toInt()
                                                isLooping = true
                                                println("setting counter to " + action.data.value)
                                            }
                                            loopStartIndex = curr - 1
                                        }
                                        // If loop end encountered, jump back if counter still > 1
                                        else if (action.name == "LOOP END")
                                        {
                                            if (loopCounter > 1 && isLooping)
                                            {
                                                loopCounter--
                                                curr = loopStartIndex
                                                println("should jump back to " + curr )
                                            }
                                            else
                                            {
                                                isLooping = false
                                                println("loopcounter == 0 or isLooping")
                                            }
                                        }

                                        // If action goes backward, break
                                        if ((((curr - 1) * stepvalue) <  scrollState.value) && !isLooping)
                                        {
                                            userStates = "STOP"
                                            isStop = false
                                            isPlay = false
                                            isPause = false
                                            isLooping = false
                                            println("Breaking out")
                                            break
                                        }
                                        // Move scroll forward
                                        scrollState.animateScrollTo((curr * stepvalue), animationSpec = tween(durationMillis = 1000, easing = LinearEasing))
                                        curr = (scrollState.value/stepvalue) + 1
                                        remainder = scrollState.value % stepvalue
                                        Log.d("ActionIndex", "Current Action: ${(scrollState.value/stepvalue) + 1}")
                                    }
                                }
                                // Reset states after loop ends
                                resetAllActionDuration(actionsInSequenceBar)
                                playPause = false
                                isPlay = false
                                isPause = false
                                userStates = "STOP"
                                updateSequenceBarColor(defaultColor)
                                updateScrollEnabled(true)
                            }
                        }
                    },
                    modifier = Modifier.size(70.dp)
                ) {
                    Icon(
                        imageVector = icon,
                        contentDescription = if (playPause) "Pause" else "Play",
                        tint = if (!playPause) playButtonColor else Color.Blue,
                        modifier = Modifier.fillMaxSize()
                    )
                }
            }
        }
    }
}

/**
 * Executes an action by adding it to the ROS action list (simulation).
 *
 * @param action The action to execute.
 * @param viewModel The ViewModel managing the connection and communication with AIDA.
 */
private fun executeAction(
    action: BaseAction,
    viewModel: MainViewModel
) {
    // Sends an action to ROS part of AIDA for execution
    if (viewModel.isSequenceConnected) {
        viewModel.sendActionsToExecute(mutableListOf(action)) // Send action to AIDA for execution
        println("[Executing action] >> ${action.name} | Data: ${action.data.value}")

    }
}


/**
 * Adds an action to the sequence bar.
 *
 * If the action is "LOOP", it actually adds a LOOP START and LOOP END pair.
 * Otherwise, it copies the action's current state and adds it.
 *
 * @param action The action to add.
 * @param actionsInSequenceBar The mutable list that represents the current sequence.
 * @param scrollState The scroll state of the sequence bar.
 * @param coroutineScope A scope to launch scrolling animations if needed.
 */
private fun addActionToSequenceBar(action: BaseAction, actionsInSequenceBar: MutableList<BaseAction>, scrollState: ScrollState, coroutineScope: CoroutineScope)
{

    // If many actions are present, scroll to the end when adding a new one
    if(actionsInSequenceBar.size > 3)
    {
        coroutineScope.launch {
            scrollState.animateScrollTo((344 * (actionsInSequenceBar.size)), animationSpec = tween(durationMillis = 750, easing = FastOutSlowInEasing))
        }
    }

    if (action.name == "LOOP") {
        // For LOOP, add LOOP START and LOOP END together
        // TODO: LOOP setting action settings are defind in two seperet areas HERE and in ActionsData.kt
        val loopStart = SpecialAction(id = 2, name = "LOOP START", isActive = mutableStateOf(true), isReorderable = true, data = mutableStateOf("1"), icon = mutableStateOf(Icons.Outlined.Repeat), cardColor = mutableStateOf(specialActionColor), duration = 0f, actionScale = mutableDoubleStateOf(1.0))
        val loopEnd = SpecialAction(id = 2, name = "LOOP END", isActive = mutableStateOf(true), isReorderable = true, data = mutableStateOf(""), icon = mutableStateOf(Icons.Outlined.Repeat), cardColor = mutableStateOf(specialActionColor), duration = 0f, actionScale = mutableDoubleStateOf(1.0))

        val copiedLoopStart = loopStart.clone()
        val copiedLoopEnd = loopEnd.clone()

        actionsInSequenceBar.add(copiedLoopStart)
        actionsInSequenceBar.add(copiedLoopEnd)
    }
    else
    {
        // For normal actions, create a copy with fresh mutable states so each action is independent
        val copiedAction = when (action) {
            is MoveAction -> action.clone()
            is SpecialAction -> action.clone()
            else -> action
        }

        actionsInSequenceBar.add(copiedAction)
    }
    printCurrentActions(actionsInSequenceBar)
}

private fun resetAllActionDuration(actionsInSequenceBar: List<BaseAction>) {
    for (action in actionsInSequenceBar) {
        action.remainingDuration.value = action.duration
    }
}

private fun resetActionDurationInRange(actionsInSequenceBar: List<BaseAction>, begin: Int, end: Int) {
    actionsInSequenceBar.subList(begin, end).forEach {
        it.remainingDuration.value = it.duration
    }
}

/**
 * Prints the current list of actions in the sequence bar to the console for debugging.
 */
fun printCurrentActions(actionsInSequenceBar: List<BaseAction>)
{
    val actionDetails = actionsInSequenceBar.withIndex().joinToString("\n") { (index, action) ->
        "Action: ${index + 1} | ${action.name} | ${action.data.value}"
    }
    println("============ Current actions in sequence bar: [${actionsInSequenceBar.size}] ============\n$actionDetails")
}

/**
 * Prints the action library to the console for debugging.
 */
private fun printActionLibrary() {
    val libraryActions = ActionLibraryData.withIndex().joinToString("\n") { (index, action) ->
        "Action: ${index + 1} | ${action.name} | ${action.data.value}"
    }
    println("============ Action Library  ============\n$libraryActions")
}






private suspend fun executeActionWithCountdown(
    action: BaseAction,
    viewModel: MainViewModel,
    scope: CoroutineScope
) {
    // Reset duration at start of execution
    action.remainingDuration.value = action.duration
    action.width.value = 210.dp
    action.height.value = 200.dp
    action.actionScale.value = 1.6

    // Start countdown in parallel with action execution
    val thred = scope.launch {
        while (action.remainingDuration.value > 0) {
            delay(100) // Update every 100ms
            action.remainingDuration.value = (action.remainingDuration.value - 0.1f)
                .coerceAtLeast(0f) // Prevent negative values
        }
    }

    // Execute the action
    if (viewModel.isSequenceConnected) {
        viewModel.sendActionsToExecute(mutableListOf(action))
        println("[Executing action] >> ${action.name} | Data: ${action.data.value}")
    }

    // Wait for the full duration
    //delay((action.duration * 1000).toLong())
    thred.join()
    //delay(1000)
    action.width.value = 160.dp
    action.height.value = 150.dp
    action.actionScale.value = 1.0

}