package com.example.aida.popups

import androidx.compose.animation.core.tween
import androidx.compose.animation.animateColorAsState
import androidx.compose.foundation.background
import androidx.compose.foundation.clickable
import androidx.compose.foundation.layout.*
import androidx.compose.foundation.lazy.grid.GridCells
import androidx.compose.foundation.lazy.grid.LazyVerticalGrid
import androidx.compose.foundation.lazy.grid.items
import androidx.compose.foundation.shape.RoundedCornerShape
import androidx.compose.material3.*
import androidx.compose.runtime.*
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.draw.clip
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.graphics.vector.ImageVector
import androidx.compose.ui.text.font.FontWeight
import androidx.compose.ui.text.style.TextAlign
import androidx.compose.ui.unit.dp
import androidx.compose.ui.unit.sp
import androidx.compose.ui.window.Dialog
import androidx.compose.ui.graphics.graphicsLayer

import com.example.aida.ui.composables.sequenceTabComposables.moveActionColor
import com.example.aida.ui.composables.sequenceTabComposables.specialActionColor

@Composable
fun PopupGestures(
    onDismiss: () -> Unit,
    onSave: (String, ImageVector) -> Unit
) {
    var selectedGestureName by remember { mutableStateOf<String?>(null) }
    var selectedGestureIcon by remember { mutableStateOf<ImageVector?>(null) }

    val gestureList = GestureData.toList()

    Dialog(
        onDismissRequest = onDismiss,
    ) {
        // Wraps the gesture selection content.
        Surface(
            modifier = Modifier
                .fillMaxWidth()
                .wrapContentHeight(),
            shape = RoundedCornerShape(20.dp),
            shadowElevation = 8.dp,
            tonalElevation = 2.dp,
        ) {
            Box(
                modifier = Modifier
                    .background(moveActionColor)
                    .padding(20.dp)
                    .clip(RoundedCornerShape(20.dp)),
                contentAlignment = Alignment.Center
            ){
                // Arranges the content vertically with centered alignment and padding.
                Column(
                    modifier = Modifier
                        .background(Color.White)
                        .padding(16.dp),
                    horizontalAlignment = Alignment.CenterHorizontally
                ) {
                    // Displays the title "Select a Gesture" in a large font style.
                    Text(
                        text = "Select a Gesture",
                        style = MaterialTheme.typography.titleLarge,
                        modifier = Modifier.padding(top = 8.dp, bottom = 8.dp)
                    )

                    // Displays the gesture items in a grid layout with 3 columns.
                    LazyVerticalGrid(
                        columns = GridCells.Fixed(3),
                        verticalArrangement = Arrangement.spacedBy(8.dp),
                        horizontalArrangement = Arrangement.spacedBy(8.dp),
                        modifier = Modifier
                            .padding(16.dp)
                            .background(Color.White)
                    ) {
                        // Displays gestures as GestureButton with icon and name based on gestureList.
                        items(gestureList) { gesture ->
                            GestureButton(
                                item = gesture,
                                isSelected = selectedGestureName == gesture.name && selectedGestureIcon == gesture.icon,
                                onClick = {
                                    selectedGestureName = gesture.name
                                    selectedGestureIcon = gesture.icon
                                }
                            )
                        }
                    }

                    // Container for cancel and save buttons.
                    Row(
                        horizontalArrangement = Arrangement.SpaceBetween,
                        modifier = Modifier.fillMaxWidth()
                    ) {
                        // Cancel button that closes the popup without saving anything.
                        Button(
                            onClick = onDismiss,
                            colors = ButtonDefaults.buttonColors(
                                containerColor = cancelButtonColor
                            )
                        ) {
                            Text("Cancel")
                        }

                        // Save button that saves the selected gesture and closes the popup, only enabled if a gesture is selected.
                        Button(
                            modifier = Modifier,
                            colors = ButtonDefaults.buttonColors(saveButtonColor),
                            onClick = {
                                selectedGestureName?.let { name ->
                                    selectedGestureIcon?.let { icon ->
                                        onSave(name, icon)
                                    }
                                }
                                onDismiss()
                            },
                            enabled = selectedGestureName != null && selectedGestureIcon != null
                        ) {
                            Text("Save")
                        }
                    }
                }
            }
        }
    }
}

@Composable
fun GestureButton(
    item: Gesture,
    isSelected: Boolean,
    onClick: () -> Unit
) {
    val backgroundColor by animateColorAsState(
        targetValue = if (isSelected) specialActionColor else moveActionColor,
        animationSpec = tween(durationMillis = 150),
        label = ""
    )

    // A container for the gesture button with a background color that changes based on selection state.
    Box(
        modifier = Modifier
            .fillMaxWidth()
            .height(150.dp)
            .clip(RoundedCornerShape(12.dp))
            .background(backgroundColor)
            .clickable(onClick = onClick)
    ) {
        // Arranges the icon and name vertically within the gesture button.
        Column(
            modifier = Modifier
                .fillMaxSize()
                .padding(10.dp),
            horizontalAlignment = Alignment.CenterHorizontally,
            verticalArrangement = Arrangement.SpaceEvenly
        ) {
            // Displays the icon for the gesture.
            Icon(
                imageVector = item.icon,
                contentDescription = "",
                tint = Color.White,
                modifier = Modifier
                    .size(70.dp)
                    .graphicsLayer(rotationZ = item.iconRotation)
            )
            // Displays the name of the gesture.
            Text(
                text = item.name,
                color = Color.White,
                textAlign = TextAlign.Center,
                fontWeight = FontWeight.Bold,
                fontSize = 20.sp
            )
        }
    }
}
