package com.example.aida.popups

import androidx.compose.foundation.background
import androidx.compose.foundation.clickable
import androidx.compose.foundation.layout.*
import androidx.compose.foundation.lazy.LazyRow
import androidx.compose.foundation.lazy.items
import androidx.compose.foundation.lazy.rememberLazyListState
import androidx.compose.foundation.shape.RoundedCornerShape
import androidx.compose.material3.*
import androidx.compose.runtime.*
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.draw.clip
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.text.style.TextAlign
import androidx.compose.ui.unit.dp
import androidx.compose.ui.unit.sp
import androidx.compose.ui.text.TextStyle
import androidx.compose.ui.window.Dialog
import androidx.compose.foundation.border
import androidx.compose.ui.platform.LocalDensity
import kotlin.math.abs


import com.chargemap.compose.numberpicker.NumberPicker
import com.example.aida.ui.composables.sequenceTabComposables.moveActionColor
import com.example.aida.ui.composables.sequenceTabComposables.specialActionColor

@Composable
fun PopupLoop(
    onDismiss: () -> Unit,
    onSave: (String) -> Unit
) {
    var pickerValue by remember { mutableIntStateOf(1) }

    Dialog(
        onDismissRequest = onDismiss,
    ) {
        // Wraps the loop selection content.
        Surface(
            modifier = Modifier
                .fillMaxWidth()
                .wrapContentHeight(),
            shape = RoundedCornerShape(20.dp),
            shadowElevation = 8.dp, // Adds shadow
            tonalElevation = 2.dp,
        ) {
            Box(
                modifier = Modifier
                    .background(moveActionColor)
                    .padding(20.dp)
                    .clip(RoundedCornerShape(20.dp)),
                contentAlignment = Alignment.Center
            ){
                Column(
                    modifier = Modifier
                        .background(Color.White)
                        .padding(16.dp),
                    horizontalAlignment = Alignment.CenterHorizontally
                ) {
                    Text(
                        text = "Select a loop interval",
                        style = MaterialTheme.typography.titleLarge,
                        modifier = Modifier.padding(top = 8.dp, bottom = 8.dp)
                    )

                    // Displays the number picker for selecting a loop interval.
                    NumberPicker(
                        value = pickerValue,
                        range = 1 .. 10,
                        onValueChange = {
                            pickerValue = it
                        },
                        textStyle = TextStyle(fontSize = 24.sp),
//                        modifier = Modifier
//                            .fillMaxSize()
                    )

                    Spacer(modifier = Modifier.height(16.dp))

                    Row(
                        horizontalArrangement = Arrangement.SpaceBetween,
                        modifier = Modifier.fillMaxWidth()
                    ) {
                        // Cancel button that closes the popup without saving.
                        Button(
                            onClick = onDismiss,
                            colors = ButtonDefaults.buttonColors(
                                containerColor = cancelButtonColor
                            )
                        ) {
                            Text("Cancel")
                        }

                        // Save button that saves the selected loop interval and closes the popup.
                        Button(
                            modifier = Modifier,
                            colors = ButtonDefaults.buttonColors(saveButtonColor),
                            onClick = {
                                pickerValue.let {
                                    val stringValue = it.toString()
                                    onSave(stringValue)
                                }
                                onDismiss()
                            }
                        ) {
                            Text("Save")
                        }
                    }
                }
            }
        }
    }
}

@Composable
fun HorizontalNumberPicker(
    value: Int,
    range: IntRange,
    onValueChange: (Int) -> Unit
) {
    val adjustedValue = value.coerceIn(range)
    val listState = rememberLazyListState(initialFirstVisibleItemIndex = (adjustedValue - range.first) + 1)


    // Padding value for the LazyRow
    val density = LocalDensity.current
    val paddingPx = with(density) { 30.dp.toPx() }

    // Calculate centered item index based on scroll offset and padding
    val visibleItemIndex = remember {
        derivedStateOf {
            val centerOffset = (listState.layoutInfo.viewportEndOffset - listState.layoutInfo.viewportStartOffset) / 2

            val minItem = listState.layoutInfo.visibleItemsInfo.minByOrNull {
                val absValue = abs((it.offset + it.size / 2) - centerOffset)
                println("Item Offset: ${it.offset}, Item Size: ${it.size}, Center Offset: $centerOffset, Abs Value: $absValue")
                absValue
            }

            minItem?.index?.plus(listState.firstVisibleItemIndex) ?: 0
        }
    }

    // Print the scroll values for debugging purposes
    LaunchedEffect(listState.firstVisibleItemIndex, listState.firstVisibleItemScrollOffset) {
        println("First Visible Item Index: ${listState.firstVisibleItemIndex + 1}")
        println("First Visible Item Scroll Offset: ${listState.firstVisibleItemScrollOffset}")
    }

    // Update value when the centered item changes
    LaunchedEffect(visibleItemIndex.value) {
        val centeredValue = range.elementAtOrNull(visibleItemIndex.value)
        if (centeredValue != null && centeredValue != adjustedValue) {
            onValueChange(centeredValue)
        }
    }

    Box(
        modifier = Modifier
            .fillMaxWidth()
            .height(140.dp)
    ) {
        // Number picker items
        LazyRow(
            state = listState,
            horizontalArrangement = Arrangement.spacedBy(8.dp),
            verticalAlignment = Alignment.CenterVertically,
            contentPadding = PaddingValues(horizontal = 229.dp),
            modifier = Modifier.matchParentSize()
        ) {
            items(range.toList()) { number ->
                Box(
                    contentAlignment = Alignment.Center,
                    modifier = Modifier
                        .size(50.dp)
                        .background(
                            color = moveActionColor,
                            shape = RoundedCornerShape(8.dp)
                        )
                ) {
                    Text(
                        text = number.toString(),
                        style = MaterialTheme.typography.bodyMedium.copy(
                            color = Color.White
                        ),
                        textAlign = TextAlign.Center
                    )
                }
            }
        }

        // Centered selection box
        Box(
            modifier = Modifier
                .align(Alignment.Center)
                .size(60.dp)
                .border(2.dp, specialActionColor, RoundedCornerShape(8.dp))
        )
    }
}


