package com.example.aida.socketcommunication

import com.example.aida.ui.composables.sequenceTabComposables.BaseAction
import java.nio.ByteBuffer


/**
 * Client class that connects to server and sends messages
 * containing list of actions to execute.
 */
class SequenceClient (ip : String = "localhost", port : Int = 12345, timeToTimeout : Int = 60000) : AbstractClient(ip, port, timeToTimeout){

    // List of all valid action names
    private val listOfActionNames = listOf<String>("TURN LEFT", "FORWARD", "FORWARD LONG",
        "TURN RIGHT", "PAD", "INPUT GESTURE", "INPUT VOICE", "TURN LEFT LONG", "BACKWARDS",
        "BACKWARDS LONG", "TURN RIGHT LONG", "LOOP START", "LOOP END", "INPUT SOUND")

    // List of all special action names
    private val listOfSpecialActionNames = listOf<String>("INPUT GESTURE", "INPUT VOICE",
        "INPUT SOUND", "LOOP START")

    // Map of all action names and their respective ID. Is really only defined and decided here.
    private val actionNamesMap = mapOf(Pair("TURN LEFT", 1),
        Pair("FORWARD", 2), Pair("FORWARD LONG", 3), Pair("TURN RIGHT", 4), Pair("PAD", 5),
        Pair("INPUT GESTURE", 6), Pair("INPUT VOICE", 7), Pair("TURN LEFT LONG", 8),
        Pair("BACKWARDS", 9), Pair("BACKWARDS LONG", 10), Pair("TURN RIGHT LONG", 11),
        Pair("LOOP START", 12), Pair("LOOP END", 13), Pair("INPUT SOUND", 14))



    /**
     * What we send to AIDA (message body):
     *
     * First we prepare a number (short) that represents the ON/OFF signal.
     * If ON (1), we have sent 1 or more actions for AIDA to immediately execute.
     * If OFF (2), we have sent a simple stop signal that is supposed to tell AIDA
     * to stop executing any actions right now.
     *
     *
     * After that, for each action in the list we prepare the following (done in prepareAction function):
     * First an ID (short) representing the action.
     * If it was a move action, that's it. We send the next action ID.
     *
     * If it was a special action however, we send a few extras.
     * First we also add a size (short) for the following data (string in UTF8 format) that is sent.
     * After that we lastly add the special action data (string in UTF8).
     * Then we are done with the special action and prepare the next action ID.
     * Repeat until entire list is prepared and then send to server.
    */
    /**
     * Send data containing a list of actions to execute to server
     * @param actions The list of actions to be sent to AIDA for execution
     */
    fun sendActionsToExecute(actions : MutableList<BaseAction>) {
        var size = 2 // ON/OFF signal takes 2 bytes (short)
        for (action in actions) {
            size += 2 // Short is 2 bytes and we need one short per action for ID
            if (action.name in listOfSpecialActionNames) {
                val utf8datasize = action.data.value.toByteArray().size
                size += 2 // 2 byte short allocated for saying how long the incoming string in UTF8 containing specialaction data is
                size += utf8datasize.toShort()
            }
        }
        val buffer = ByteBuffer.allocate(size)
        buffer.putShort(Instructions.ON.value) // Add on signal

        for (action in actions) {
            prepareAction(action, buffer) // Prepare each action
        }
        val id = MessageType.SEQUENCE.value
        println("Sending sequence containing following actions to server: ")
        for (action in actions) {
            println(action.name)
        }
        sendDataToServer(id, buffer.array())
    }


    /**
     * Converts an action to acceptable format for sending over the socket.
     * See the documentation for sendActionsToExecute function for more info on what this function does.
     * @param action The action to be prepared
     */
    private fun prepareAction(action : BaseAction, buffer: ByteBuffer) {
        try {
            checkIfActionNameValid(action.name)
        } catch (e : IllegalArgumentException){
            println(e.message)
            throw IllegalArgumentException("The action name to be sent is not valid! Invalid action name: $action.name")
        }

        val actionID = actionNamesMap.get(action.name)!!.toShort()
        buffer.putShort(actionID)

        if (action.name in listOfSpecialActionNames) {
            buffer.putShort(action.data.value.toByteArray().size.toShort()) // Add size of following data string in UTF8

            buffer.put(action.data.value.toByteArray()) // Add data string in UTF8
        }
    }

    /**
     * Send a signal to AIDA that tells it to stop executing any currently running actions
     */
    fun sendStopExecutingActions() {
        val size = 2
        val buffer = ByteBuffer.allocate(size)
        buffer.putShort(Instructions.OFF.value)

        val id = MessageType.SEQUENCE.value
        println("Sending a stop executing current actions signal")
        sendDataToServer(id, buffer.array())
    }

    /**
     * Check if the action name to be sent is valid
     * @param actionName The name of the action AIDA should execute
     * @throws IllegalArgumentException if the action name is not valid
     */
    private fun checkIfActionNameValid(actionName: String){
        if (actionName !in listOfActionNames) {
            throw IllegalArgumentException("The action name to be sent is not valid! Invalid action name: $actionName")
        }
    }
}