package com.example.aida.socketcommunication
import java.io.BufferedReader
import java.io.File
import java.io.InputStreamReader
import java.net.ServerSocket
import java.nio.ByteBuffer
import java.nio.ByteOrder

/**
 * Server class that listens for incoming connections
 * and sends data to the client
 * -- For testing purposes, not used in application
 */

class Server (port: Int = 12345) {
    private val serverSocket = ServerSocket(port)
    private val socket = serverSocket.accept()

    /**
     * Send data to the client
     * @param id The id of the message
     * @param data The data to be sent
     * For testing purposes
     */
    fun sendData(id: Short, data: ByteArray) {
        //val id = MessageType.STT.value
        val frameSize = data.size
        val frameBufferSize = ByteBuffer.allocate(6)
        frameBufferSize.putShort(id)
        frameBufferSize.putInt(frameSize)

        frameBufferSize.order(ByteOrder.BIG_ENDIAN)
        socket.getOutputStream().write(frameBufferSize.array())
        socket.getOutputStream().write(data)
    }

    /**
     * Close the connection to the socket
     *
     */
    fun closeConnection() {
        socket.close()
        serverSocket.close()
    }

    /**
     * Get the header of the message sent over socket
     * @return Pair with id of the message and size of message
     * For testing purposes
     */
    fun getHeader() : Pair<Short, Int>{
        var headerBuffer = ByteBuffer.allocate(6)
        socket.getInputStream().read(headerBuffer.array())
        headerBuffer = ByteBuffer.wrap(headerBuffer.array()).order(ByteOrder.BIG_ENDIAN)
        val id  = headerBuffer.short
        val size = headerBuffer.int

        return Pair(id, size)
    }

    /**
     * Get the body of message sent over socket
     * @return The message as a ByteArray
     * For testing purposes
     */
    fun getBody(size : Int) : ByteArray {
        val frameBuffer = ByteBuffer.allocate(size)
        var bytesRead = 0

        while (bytesRead < size) {
            bytesRead += socket.getInputStream().read(frameBuffer.array(), bytesRead, size - bytesRead)
        }
        return frameBuffer.array()
    }

    /**
     * Only for debug
     */
    fun receiveMessage() {
        val (id, size) = getHeader()
        val message = getBody(size)
        //if(id == MessageType.REQ_STT.value) {
        println("Message received: ${String(message)}")
        //}
        //sendData(id, "Hello, World!".toByteArray())
    }

    /**
     * The following function decodes a sequence message sent over the socket.
     * Made for testing purposes but also highlights how the decoding can be done on the AIDA
     * @param message the body of the sequence message sent over socket
     */
    fun decodeSequenceMessage(message: ByteArray) {
        val messageBufferShortOnOff = ByteBuffer.wrap(message.copyOfRange(0, 2)).asShortBuffer() // Extract the ON/OFF signal from message

        println("Server received message ON/OFF signal: " + messageBufferShortOnOff.get(0))
        if (messageBufferShortOnOff.get(0).toInt() == 1) { // Is it an ON signal? AKA should we execute actions?
            var i = 2 // ON/OFF signal takes up byte 0-1 so actionIDs start on index 2
            while (i < message.size) {
                val actionID = ByteBuffer.wrap(message.copyOfRange(i, i+2)).asShortBuffer().get(0) // Extract the ID of an action in the message to figure out which action should be executed
                println("Server currently processing following action ID: $actionID")

                if (actionID.toInt() == 6 || actionID.toInt() == 7 || actionID.toInt() == 12 || actionID.toInt() == 14) { // Checks if it is a special action
                    val specialDataSize = ByteBuffer.wrap(message.copyOfRange(i+2, i+4)).asShortBuffer().get(0) // Extract the size of the string data in the special action
                    val specialDataString = message.decodeToString(i+4, i+4+specialDataSize) // Extract the string data in the special action
                    println("Special Action with ID: $actionID has the following data string: $specialDataString")
                    i += 4 + specialDataSize // Move i forward to next action in the ByteArray
                }
                else {
                    i += 2
                }
            }
        }
    }
}