package com.example.aida.ui.composables.sequenceTabComposables

import androidx.compose.runtime.MutableState
import androidx.compose.runtime.mutableStateListOf
import androidx.compose.runtime.mutableStateOf
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.outlined.*
import androidx.compose.runtime.mutableDoubleStateOf
import androidx.compose.runtime.mutableFloatStateOf
import androidx.compose.ui.graphics.vector.ImageVector
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.unit.Dp
import androidx.compose.ui.unit.dp

/**
 * Defines a set of colors used for action components.
 */
val moveActionColor = Color(0xff6c779c)    // Color for move actions
val specialActionColor = Color(0xff435159) // Color for special actions
val activeColor = Color(0xFF5B7A8C)        // Color indicating an active state
val inActiveColor = Color(0xff8A8A8A)      // Color indicating an inactive state
/**
 * Represents the base attributes of an action.
 *
 * Note:
 *   - `isActive` determines whether the action is active on the sequence bar.
 *   -  Actions become active when assigned a data value:
 *      - `MoveAction` instances are active by default.
 *      - `SpecialAction` instances are inactive by default, except for the "Loop" action, which starts as active.
 */
interface BaseAction: Cloneable {
    val id: Int                         // A unique identifier for each action
    val name: String                    // The name of the action
    var duration: Float // Expected duration of action
    var remainingDuration: MutableState<Float>
    var isActive: MutableState<Boolean> // Indicates whether the action is active on the sequence bar, the action becomes active when assigned a data value
    var isReorderable: Boolean          // Determines if the action can be reordered
    var data: MutableState<String>      // Stores data related to the action
    var icon: MutableState<ImageVector> // The icon displayed for the action
    var cardColor: MutableState<Color> // The color of the action's visual container
    var height: MutableState<Dp>
    var width: MutableState<Dp>
    var actionScale: MutableState<Double>

    public override fun clone(): BaseAction
    }
/**
 * These data classes implement `BaseAction` and can define their own specific attributes.
 */
data class MoveAction(
    override val id: Int,
    override val name: String,
    override var isActive: MutableState<Boolean> = mutableStateOf(true),      // Default value, used unless explicitly overridden
    override var isReorderable: Boolean,
    override var duration: Float,
    override var remainingDuration: MutableState<Float> = mutableFloatStateOf(duration),
    override var data: MutableState<String>,
    override var icon: MutableState<ImageVector>,
    override var cardColor: MutableState<Color> = mutableStateOf(moveActionColor),   // Default value, used unless explicitly overridden
    override var height: MutableState<Dp> = mutableStateOf(150.dp),
    override var width: MutableState<Dp> = mutableStateOf(160.dp),
    override var actionScale: MutableState<Double> = mutableDoubleStateOf(1.0)
) : BaseAction {

    override fun clone(): MoveAction {
        return MoveAction(
            id = this.id,
            name = this.name,
            isActive = mutableStateOf(this.isActive.value),
            isReorderable = this.isReorderable,
            duration = this.duration,
            remainingDuration = mutableFloatStateOf(this.remainingDuration.value),
            data = mutableStateOf(this.data.value),
            icon = mutableStateOf(this.icon.value),
            cardColor = mutableStateOf(this.cardColor.value),
            height = mutableStateOf(this.height.value),
            width = mutableStateOf(this.width.value),
            actionScale = mutableDoubleStateOf(this.actionScale.value)
        )
    }
}

data class SpecialAction(
    override val id: Int,
    override val name: String,
    override var duration: Float,
    override var remainingDuration: MutableState<Float> = mutableFloatStateOf(duration),
    override var isActive: MutableState<Boolean> = mutableStateOf(false),       // Default value, used unless explicitly overridden
    override var isReorderable: Boolean,
    override var data: MutableState<String>,
    override var icon: MutableState<ImageVector>,
    override var cardColor: MutableState<Color> = mutableStateOf(specialActionColor), // Default value, used unless explicitly overridden
    override var height: MutableState<Dp> = mutableStateOf(150.dp),
    override var width: MutableState<Dp> = mutableStateOf(160.dp),
    override var actionScale: MutableState<Double> = mutableDoubleStateOf(1.0),
    val iconRotation: MutableState<Float> = mutableStateOf(0f)
) : BaseAction {

    override fun clone(): SpecialAction {
        return SpecialAction(
            id = this.id,
            name = this.name,
            duration = this.duration,
            remainingDuration = mutableFloatStateOf(this.remainingDuration.value),
            isActive = mutableStateOf(this.isActive.value), 
            isReorderable = this.isReorderable,
            data = mutableStateOf(this.data.value), 
            icon = mutableStateOf(this.icon.value), 
            cardColor = mutableStateOf(this.cardColor.value), 
            height = mutableStateOf(this.height.value), 
            width = mutableStateOf(this.width.value),  
            actionScale = mutableDoubleStateOf(this.actionScale.value),  
            iconRotation = mutableStateOf(this.iconRotation.value) 
        )
    }
}


/**
 * A predefined set of actions for the Action Library.
 *
 * Guidelines:
 * - Actions with default parameter values do not need explicit declarations in the list unless those defaults are being overridden.
 * - The `ActionLibraryData` includes both `MoveAction` and `SpecialAction` types, each serving distinct purposes in the action library.
 *
 * Note:
 * - Be cautious when assigning duplicate IDs (especially for PAD blocks) as it may lead to unexpected behavior in the app.
 */


val ActionLibraryData: Set<BaseAction> = setOf(
    MoveAction(id = 1, name = "TURN LEFT", isReorderable = true, data = mutableStateOf(""), icon = mutableStateOf(Icons.Outlined.ArrowBack), duration = 2.0f),
    MoveAction(id = 1, name = "FORWARD", isReorderable = true, data = mutableStateOf(""), icon = mutableStateOf(Icons.Outlined.ArrowUpward), duration = 2.0f),
    MoveAction(id = 1, name = "FORWARD LONG", isReorderable = true, data = mutableStateOf(""), icon = mutableStateOf(Icons.Outlined.North), duration = 5.0f),
    MoveAction(id = 1, name = "TURN RIGHT", isReorderable = true, data = mutableStateOf(""), icon = mutableStateOf(Icons.Outlined.ArrowForward), duration = 2.0f),
    MoveAction(id = 888, name = "PAD", isReorderable = false, data = mutableStateOf(""), icon = mutableStateOf(Icons.Outlined.Settings), duration = 0f),
    SpecialAction(id = 2, name = "INPUT GESTURE", isReorderable = true, data = mutableStateOf("EMPTY"), icon = mutableStateOf(Icons.Outlined.WavingHand), cardColor = mutableStateOf(inActiveColor), duration = 0f),
    SpecialAction(id = 2, name = "INPUT VOICE", isReorderable = true, data = mutableStateOf("EMPTY"), icon = mutableStateOf(Icons.Outlined.Campaign), cardColor = mutableStateOf(inActiveColor),duration = 0f),
    MoveAction(id = 1, name = "TURN LEFT LONG", isReorderable = true, data = mutableStateOf(""), icon = mutableStateOf(Icons.Outlined.West),duration = 5.0f),
    MoveAction(id = 1, name = "BACKWARDS", isReorderable = true, data = mutableStateOf(""), icon = mutableStateOf(Icons.Outlined.ArrowDownward),duration = 2.0f),
    MoveAction(id = 1, name = "BACKWARDS LONG", isReorderable = true, data = mutableStateOf(""), icon = mutableStateOf(Icons.Outlined.South),duration = 5.0f),
    MoveAction(id = 1, name = "TURN RIGHT LONG", isReorderable = true, data = mutableStateOf(""), icon = mutableStateOf(Icons.Outlined.East),duration = 5.0f),
    MoveAction(id = 999, name = "PAD", isReorderable = false, data = mutableStateOf(""), icon = mutableStateOf(Icons.Outlined.Settings),duration = 0f),
    SpecialAction(id = 2, name = "LOOP", isActive = mutableStateOf(true), isReorderable = false, data = mutableStateOf(""), icon = mutableStateOf(Icons.Outlined.Repeat),duration = 0f),
    SpecialAction(id = 2, name = "INPUT SOUND", isReorderable = true, data = mutableStateOf("EMPTY"), icon = mutableStateOf(Icons.Outlined.MusicNote), cardColor = mutableStateOf(inActiveColor),duration = 0f)
)




