package com.example.aida.ui.composables.sequenceTabComposables

import android.util.Log
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.launch
import sh.calvin.reorderable.ReorderableRow
import androidx.compose.animation.*
import androidx.compose.animation.core.EaseInOut
import androidx.compose.animation.core.FastOutSlowInEasing
import androidx.compose.animation.core.Spring
import androidx.compose.animation.core.spring
import androidx.compose.animation.core.tween
import androidx.compose.foundation.Canvas
import androidx.compose.foundation.ScrollState
import androidx.compose.foundation.background
import androidx.compose.foundation.border
import androidx.compose.foundation.clickable
import androidx.compose.foundation.gestures.DraggableState
import androidx.compose.foundation.gestures.Orientation
import androidx.compose.foundation.horizontalScroll
import androidx.compose.foundation.interaction.MutableInteractionSource
import androidx.compose.foundation.layout.*
import androidx.compose.foundation.shape.RoundedCornerShape
import androidx.compose.foundation.gestures.detectTapGestures
import androidx.compose.foundation.gestures.draggable
import androidx.compose.foundation.gestures.rememberDraggableState
import androidx.compose.foundation.gestures.scrollBy
import androidx.compose.foundation.layout.Box
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.filled.Close
import androidx.compose.material.icons.filled.Settings
import androidx.compose.material3.Card
import androidx.compose.material3.CardDefaults
import androidx.compose.material3.Icon
import androidx.compose.material3.IconButton
import androidx.compose.material3.Text
import androidx.compose.material3.Button
import androidx.compose.runtime.Composable
import androidx.compose.runtime.mutableStateOf
import androidx.compose.runtime.mutableIntStateOf
import androidx.compose.runtime.key
import androidx.compose.runtime.mutableLongStateOf
import androidx.compose.runtime.remember
import androidx.compose.runtime.snapshots.SnapshotStateList
import androidx.compose.runtime.getValue
import androidx.compose.runtime.mutableFloatStateOf
import androidx.compose.runtime.rememberCoroutineScope
import androidx.compose.runtime.setValue
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.draw.clip
import androidx.compose.ui.geometry.CornerRadius
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.graphics.graphicsLayer
import androidx.compose.ui.hapticfeedback.HapticFeedbackType
import androidx.compose.ui.platform.LocalDensity
import androidx.compose.ui.platform.LocalHapticFeedback
import androidx.compose.ui.semantics.CustomAccessibilityAction
import androidx.compose.ui.semantics.customActions
import androidx.compose.ui.semantics.semantics
import androidx.compose.ui.text.TextStyle
import androidx.compose.ui.text.font.FontWeight
import androidx.compose.ui.text.style.TextAlign
import androidx.compose.ui.unit.dp
import androidx.compose.ui.unit.sp
import androidx.compose.ui.zIndex
import androidx.compose.ui.input.pointer.pointerInput
import androidx.compose.ui.graphics.vector.ImageVector

import com.example.aida.popups.PopupGestures
import com.example.aida.popups.PopupVoices
import com.example.aida.popups.PopupSounds
import com.example.aida.popups.PopupLoop
import androidx.compose.ui.platform.LocalContext
import androidx.compose.ui.unit.Dp
import androidx.compose.ui.unit.IntOffset
import com.example.aida.viewmodels.MainViewModel
import java.util.Locale
import kotlin.math.roundToInt

/**
 * Displays a horizontal "sequence bar" where users can arrange and configure actions.
 *
 * This composable shows a row of actions that can be reordered (if allowed by the item (loops do not permit this)),
 * edited via double-tap or a settings button, and removed. It also includes a scrollable timeline
 * and a progress bar to indicate the current position when scrolling horizontally.
 *
 * @param actionsInSequenceBar A mutable list (state) of actions displayed in the sequence bar.
 * @param sequenceBarColor The background color for the entire sequence bar area.
 * @param scrollState The [ScrollState] controlling horizontal scrolling of the bar.
 * @param coroutineScope The [CoroutineScope] used to launch animations or asynchronous tasks.
 * @param scrollEnabled Whether horizontal scrolling and reordering are enabled. If disabled, a
 *                      transparent overlay prevents manual scrolling.
 */
@Composable
fun SequenceBar(
    actionsInSequenceBar: SnapshotStateList<BaseAction>,
    sequenceBarColor : Color,
    scrollState: ScrollState,
    coroutineScope: CoroutineScope,
    scrollEnabled: Boolean,
    onPlayTTS: (String) -> Unit  // New parameter
) {
    // Provides access to the haptic feedback system
    val haptic = LocalHapticFeedback.current

    // Tracks interaction states for UI components
    val interactionSource = remember { MutableInteractionSource() }

    // Width of the scrollbar in density-independent pixels
    val totalWidth = 800.dp

    // State to toggle the visibility of a popup
    var togglePopUp by remember { mutableStateOf(false) }

    // Stores the index of the currently selected action
    var selectedActionIndex by remember { mutableIntStateOf(0) }

    // Retrieves the currently selected action from the list, or null if out of bounds
    val selectedAction = actionsInSequenceBar.getOrNull(selectedActionIndex)

    // Provides the current context for accessing resources or performing actions
    val context = LocalContext.current

    // Sequence bar main container
    Box(
        modifier = Modifier
            .fillMaxSize()
            .background(color = sequenceBarColor),
        contentAlignment = Alignment.Center
    ){

        // Show a popup if a special action has been double-clicked and "togglePopUp" is true.
        if (selectedAction is SpecialAction && togglePopUp)
        {
            when (selectedAction.name) {
                "INPUT SOUND" -> {
                    // Popup for configuring sound actions
                    PopupSounds(
                        onDismiss = { togglePopUp = false },
                        onSave = { inputString ->
                            val saveLambda = saveActionData(actionsInSequenceBar, selectedActionIndex)
                            saveLambda(inputString)
                            togglePopUp = false
                        }
                    )
                }
                "LOOP START" -> {
                    // Popup for configuring loop start actions
                    PopupLoop(
                        onDismiss = { togglePopUp = false },
                        onSave = { inputString ->
                            val saveLambda = saveActionData(actionsInSequenceBar, selectedActionIndex)
                            saveLambda(inputString)
                            togglePopUp = false
                        }
                    )
                }
                "INPUT GESTURE" -> {
                    // Popup for selecting gestures
                    PopupGestures(
                        onDismiss = { togglePopUp = false },
                        onSave = { name, icon ->
                            val saveLambda = saveActionData(actionsInSequenceBar, selectedActionIndex, icon)
                            saveLambda(name)
                            togglePopUp = false
                        }
                    )
                }
                "INPUT VOICE" -> {
                    PopupVoices(
                        onDismiss = { togglePopUp = false },
                        onSave = { text ->
                            val saveLambda = saveActionData(actionsInSequenceBar, selectedActionIndex)
                            saveLambda(text) // This sets the action's data.value to the selected text
                            togglePopUp = false
                        },
                        onPlay = { voice, text ->
                            onPlayTTS(text) // Call onPlayTTS with the text to test TTS immediately
                        },
                        context = context // Pass context here
                    )
                }
            }
        }

        if(!scrollEnabled)
        {
            // If scrolling is disabled (e.g., while playing), add a transparent overlay to block interaction.
            Box(
                modifier = Modifier
                    .align(Alignment.Center)
                    .background(Color.Transparent)
                    .fillMaxSize()
                    .zIndex(3f)
                    .pointerInput(Unit) {
                        // detectTapGestures does nothing here; it just intercepts to prevent interaction.
                        detectTapGestures { }
                    }
            ) {}
        }

        // Button to clear all actions from the sequence bar
        ClearSequenceButton(
            actionsInSequenceBar = actionsInSequenceBar,
            modifier = Modifier
                .align(Alignment.TopStart)
                .offset(x = 10.dp, y = 10.dp)
                .size(135.dp, 45.dp)
                .zIndex(2f)
        )

        // Label to display the time of all actions combined in the sequence
        Text(
            text = "Total duration: ${sequenceBarTotalDuration(actionsInSequenceBar).toString()}",
            modifier = Modifier
                .align(Alignment.TopEnd)
                .offset(x = (-10).dp, y=10.dp)

        )


        // Timeline (vertical line) on the left side to indicate the starting point
        TimeLine()

        // Progress bar at the bottom to show scrolling position relative to total width
        ProgressBar(
            scrollState = scrollState,
            totalWidth = totalWidth,
            actionsInSequenceBar = actionsInSequenceBar,
            modifier = Modifier
                .align(Alignment.BottomCenter)
                .padding(bottom = 16.dp)
                .zIndex(2f)
        )



        // Reorderable row of actions
        // Allows drag-and-drop reordering if items permit it.
        ReorderableRow(
            modifier = Modifier
                .fillMaxSize()
                .background(color = Color.Transparent)
                .horizontalScroll(scrollState, enabled = scrollEnabled)
                // Padding creates space to scroll and position items further right.
                .padding(PaddingValues(start = 560.dp, top = 8.dp, end = 720.dp, bottom = 8.dp)),
            verticalAlignment = Alignment.CenterVertically,
            horizontalArrangement = Arrangement.spacedBy(12 .dp),
            list = actionsInSequenceBar.toList(),
            onSettle = { fromIndex, toIndex ->
                if (fromIndex in actionsInSequenceBar.indices && toIndex in actionsInSequenceBar.indices) {
                    val fromItem = actionsInSequenceBar[fromIndex]
                    // Only reorder if the action allows reordering
                    if (fromItem.isReorderable) {
                        actionsInSequenceBar.add(
                            toIndex,
                            actionsInSequenceBar.removeAt(fromIndex)
                        )
                    }else {
                        Log.e("SequenceBar, onSettle", "Error: fromIndex or to Index out of bounds in onSettle.")
                    }
                }
            },
            onMove = {
                // Provide haptic feedback when dragging starts
                haptic.performHapticFeedback(HapticFeedbackType.LongPress)
            },
        ) { index, item, _ ->
            key(item.id) {
                val lastClickTime = remember { mutableLongStateOf(0L) }
                // Each action is represented as a card:

                AnimatedVisibility(
                    visible = true,
                    enter = scaleIn(),
                    exit = scaleOut()
                ) {
                    Card(
                        modifier = Modifier
                            .animateContentSize(
                                animationSpec = tween(durationMillis = 250, easing = EaseInOut),
                              alignment = Alignment.Center
                           )
                            .width(item.width.value)
                            .height(item.height.value)
                            .clip(RoundedCornerShape(12.dp))
                            .background(item.cardColor.value)
                            .clickable {
                                // Detect double-click to open the popup:
                                val currentTime = System.currentTimeMillis()
                                if (currentTime - lastClickTime.longValue < 500) {
                                    println("Action pressed: " + item::class.simpleName)
                                    togglePopUp = true
                                    selectedActionIndex = index
                                }
                                lastClickTime.longValue = currentTime
                            }
                            .semantics {
                                // Accessibility custom actions for moving items left or right
                                customActions = listOf(
                                    CustomAccessibilityAction(
                                        label = "Move Left",
                                        action = {
                                            if (index > 0) {
                                                actionsInSequenceBar.add(
                                                    index - 1,
                                                    actionsInSequenceBar.removeAt(index)
                                                )

                                                true
                                            } else {
                                                false
                                            }
                                        }
                                    ),
                                    CustomAccessibilityAction(
                                        label = "Move Right",
                                        action = {
                                            if (index < actionsInSequenceBar.size - 1) {
                                                actionsInSequenceBar.add(
                                                    index + 1,
                                                    actionsInSequenceBar.removeAt(index)
                                                )

                                                true
                                            } else {
                                                false
                                            }
                                        }
                                    ),
                                )
                            }

                            // If the item can be reordered, attach a draggable handle
                            .then(
                                if (item.isReorderable) {
                                    Modifier.draggableHandle(
                                        onDragStarted = {
                                            haptic.performHapticFeedback(HapticFeedbackType.LongPress)
                                        },
                                        onDragStopped = {
                                            haptic.performHapticFeedback(HapticFeedbackType.LongPress)
                                        },
                                        interactionSource = interactionSource
                                    )
                                } else {
                                    Modifier
                                }
                            ),
                        colors = CardDefaults.cardColors(containerColor = item.cardColor.value)
                    ) {
                        // Main container inside an action block
                        // Content inside each action's card:
                        Box(
                            modifier = Modifier
                                .fillMaxSize(),
                            contentAlignment = Alignment.Center
                        ) {
                            Column(
                                modifier = Modifier
                                    .fillMaxSize()
                                    .padding(top = 10.dp, start = 5.dp, end = 5.dp),
                                horizontalAlignment = Alignment.CenterHorizontally,
                                verticalArrangement = Arrangement.SpaceEvenly,
                            ) {
                                // Different layouts for special cases like LOOP START/END
                                if (item.name == "LOOP START" || item.name == "LOOP END") {
                                    Column(
                                        modifier = Modifier
                                            .fillMaxSize()
                                            .padding(bottom = 18.dp),
                                        horizontalAlignment = Alignment.CenterHorizontally
                                    ) {
                                        Box(
                                            modifier = Modifier
                                                .weight(1f),
                                            contentAlignment = Alignment.Center
                                        ) {
                                            // Icon and loop parameter (e.g. number of loops)

                                            Icon(
                                                imageVector = item.icon.value,
                                                contentDescription = "",
                                                tint = Color.White,
                                                modifier = Modifier
                                                    .size((80 * item.actionScale.value).dp)
                                            )
                                            if (item.name != "LOOP END") {
                                                Text(
                                                    text = minimizeText(
                                                        item.data.value,
                                                        maxLength = 15
                                                    ),
                                                    color = Color.White,
                                                    textAlign = TextAlign.Center,
                                                    fontWeight = FontWeight.Bold,
                                                    fontSize = (20 * item.actionScale.value).sp
                                                )
                                            }
                                        }
                                        Text(
                                            text = item.name,
                                            color = Color.White,
                                            fontWeight = FontWeight.Bold,
                                            style = TextStyle(
                                                fontSize = (15 * item.actionScale.value).sp,
                                                textAlign = TextAlign.Center
                                            )
                                        )
                                    }
                                }
//                            else if (item.name == "LOOP END")
//                            {
//                                Text(
//                                    text = item.name,
//                                    color = Color.White,
//                                    fontWeight = FontWeight.Bold,
//                                    style = TextStyle(
//                                        fontSize = 20.sp,
//                                        textAlign = TextAlign.Center
//                                    )
//                                )
//                            }
                                else if (item::class.simpleName == "SpecialAction") {
                                    // Display the special action icon and data (e.g., a gesture name)
                                    Icon(
                                        imageVector = item.icon.value,
                                        contentDescription = "",
                                        tint = Color.White,
                                        modifier = Modifier
                                            .size((70 * item.actionScale.value).dp)
                                            .graphicsLayer(
                                                rotationZ = if (item.data.value == "Finger gun") 90f else 0f
                                            )
                                    )
                                    Text(
                                        text = minimizeText(item.data.value, maxLength = 15),
                                        color = Color.White,
                                        textAlign = TextAlign.Center,
                                        fontWeight = FontWeight.Bold,
                                        fontSize = (20 * item.actionScale.value).sp
                                    )
                                } else {
                                    // Default case for normal actions: show icon and name
                                    Text(
                                        text = String.format(
                                            Locale.getDefault(),
                                            "%.2f",
                                            item.remainingDuration.value
                                        ),
                                        color = Color.White,
                                        fontWeight = FontWeight.Bold,
                                        textAlign = TextAlign.Start,
                                        fontSize = (15 * item.actionScale.value).sp


                                    )
                                    Icon(
                                        imageVector = item.icon.value,
                                        contentDescription = "",
                                        tint = Color.White,
                                        modifier = Modifier
                                            .size((70 * item.actionScale.value).dp)
                                    )
                                    Text(
                                        text = item.name,
                                        color = Color.White,
                                        textAlign = TextAlign.Center,
                                        fontWeight = FontWeight.Bold,
                                        fontSize = (15 * item.actionScale.value).sp
                                    )
                                }
                            }
                            // If this is a special action (except LOOP END), show a settings button to open popup
                            if (item::class.simpleName == "SpecialAction" && item.name != "LOOP END") {
                                IconButton(
                                    onClick = {
                                        println("Action pressed: " + item::class.simpleName)
                                        togglePopUp = true
                                        selectedActionIndex = index
                                    },
                                    modifier = Modifier
                                        .align(Alignment.TopStart)
                                        .padding(4.dp)
                                        .size(30.dp)
                                ) {
                                    Icon(
                                        imageVector = Icons.Filled.Settings,
                                        contentDescription = "Bold Close",
                                        tint = Color.White,
                                        modifier = Modifier.fillMaxSize()
                                    )
                                }
                            }
                            // Close (X) button to remove the action from the bar
                            IconButton(
                                onClick = {
                                    removeActionFromSequenceBar(actionsInSequenceBar, index)
                                    Log.d("SequenceBar", "Removed action: ${item.name}")

                                    // If actions are few, animate scroll back to the start
                                    if (actionsInSequenceBar.size < 5) {
                                        coroutineScope.launch {
                                            scrollState.animateScrollTo(
                                                0,
                                                animationSpec = tween(
                                                    durationMillis = 750,
                                                    easing = FastOutSlowInEasing
                                                )
                                            )
                                        }
                                    }
                                },
                                modifier = Modifier
                                    .align(Alignment.TopEnd)
                                    .padding(4.dp)
                                    .size(30.dp)
                            ) {
                                Icon(
                                    imageVector = Icons.Filled.Close,
                                    contentDescription = "Bold Close",
                                    tint = Color.White,
                                    modifier = Modifier.fillMaxSize()
                                )
                            }
                        }
                    }
                }
            }
        }
    }

    sortLoopButtons(actionsInSequenceBar)
}

/**
 * A button that clears all actions from the sequence bar.
 *
 * @param actionsInSequenceBar The list of actions to clear.
 * @param modifier Optional [Modifier] for layout or styling.
 */
@Composable
fun ClearSequenceButton(
    actionsInSequenceBar: SnapshotStateList<BaseAction>,
    modifier: Modifier = Modifier
){
    Button(
        onClick = {
            actionsInSequenceBar.clear()
            println("Cleared sequence bar")
        },
        modifier = modifier
    ) {
        Text(
            text = "Clear Sequence",
            style = TextStyle(
                fontSize = 13.sp,
                fontWeight = FontWeight.Normal
            )
        )
    }
}

/**
 * Displays a horizontal progress bar (scrollbar) indicating the user's scroll position relative to the total width of the sequence.
 *
 * @param scrollState The state representing the current scroll position.
 * @param totalWidth The total width in dp of the scrollable area.
 * @param actionsInSequenceBar The list of actions. The count influences thumb size.
 * @param modifier Optional [Modifier] for layout or styling.
 */

@Composable
fun ProgressBar(
    scrollState: ScrollState,
    totalWidth: Dp,
    actionsInSequenceBar : SnapshotStateList<BaseAction>,
    modifier: Modifier = Modifier
) {

    // scope for the coroutine
    val scope = rememberCoroutineScope()

    // Density of the device
    val density = LocalDensity.current

    // Progressbar container
    Box(
        modifier
            .width(totalWidth)
            .height(32.dp)
            .background(color = Color.LightGray, shape = RoundedCornerShape(16.dp))
            .border(
                width = 1.dp,
                color = Color.Gray,
                shape = RoundedCornerShape(16.dp)
            )
    ) {
        val thumbSize = 1f / actionsInSequenceBar.size.coerceAtLeast(1)
        val thumbPosition = scrollState.value / (scrollState.maxValue.toFloat())

        val thumbWidth = totalWidth * thumbSize
        val thumbPositionX = (totalWidth * thumbPosition).coerceIn(0.dp, totalWidth - thumbWidth)

        // Draggable thumb
        Box(
            Modifier
                .offset( x = thumbPositionX)
                .width(thumbWidth)
                .fillMaxHeight()
                .background(color = Color(0xBF3A88D1), shape = RoundedCornerShape(16.dp))
                .draggable(
                    orientation = Orientation.Horizontal,
                    state = rememberDraggableState { delta ->
                        val deltaDp = with(density) { delta.toDp() }
                        val scrollDelta = (deltaDp / totalWidth) * scrollState.maxValue

                        scope.launch {
                            scrollState.scrollBy(scrollDelta)
                        }
                    }
                )
        )
    }
}

/**
 * Displays a vertical timeline line on the left side of the sequence bar.
 * Typically used as a marker for the start position.
 */
@Composable
fun TimeLine()
{
    AnimatedVisibility(
        visible = true,
        enter = fadeIn(tween(durationMillis = 500)),
        exit = fadeOut(tween(durationMillis = 500))
    ) {
        // Green vertical line acting as a timeline indicator.
        Box(
            modifier = Modifier
                .fillMaxHeight()
                .width(160.dp)
                .background(Color(0x4D099D0B))
        )
    }
}

/**
 * Removes an action (or a loop pair) from the sequence bar.
 *
 * Special handling:
 * - Removing a LOOP START also removes its corresponding LOOP END.
 * - Removing a LOOP END also removes its corresponding LOOP START.
 * For normal actions, it just removes that action.
 *
 * @param actionsInSequenceBar The list of actions currently in the sequence.
 * @param index The index of the action to remove.
 */
private fun removeActionFromSequenceBar(actionsInSequenceBar : SnapshotStateList<BaseAction>, index : Int)
{
    val actionToRemove = actionsInSequenceBar[index].name

    // If loop start or loop end is removed, remove the corresponding loop-pair-action, otherwise just remove the action
    if (actionToRemove == "LOOP START")
    {
        // Find the corresponding LOOP END and remove both
        var nextIndex = index + 1
        var nextAction = actionsInSequenceBar.getOrNull(nextIndex)?.name

        // Search for LOOP END
        while (nextAction != "LOOP END" && nextIndex < actionsInSequenceBar.size) {
            nextIndex++
            nextAction = actionsInSequenceBar.getOrNull(nextIndex)?.name
        }

        if (nextAction == "LOOP END") {
            actionsInSequenceBar.removeAt(index)                // Remove LOOP START
            actionsInSequenceBar.removeAt(nextIndex - 1)  // Remove LOOP END
        }
    }
    else if (actionToRemove == "LOOP END")
    {
        // Find the corresponding LOOP START and remove both
        var nextIndex = index - 1
        var nextAction = actionsInSequenceBar.getOrNull(nextIndex)?.name

        // Search for LOOP START
        while (nextAction != "LOOP START" && nextIndex >= 0) {
            nextIndex--
            nextAction = actionsInSequenceBar.getOrNull(nextIndex)?.name
        }

        if (nextAction == "LOOP START") {
            actionsInSequenceBar.removeAt(index)      // Remove LOOP END
            actionsInSequenceBar.removeAt(nextIndex)  // Remove LOOP START
        }
    }
    else
    {
        actionsInSequenceBar.removeAt(index)
    }
}

/**
 * Returns a lambda function that updates the action’s data and (optionally) icon.
 *
 * @param actionsInSequenceBar The current list of actions.
 * @param selectedActionIndex The index of the action being edited.
 * @param icon An optional new icon for the action.
 * @return A function that takes a string (e.g. a new name or parameter) and updates the selected action with it.
 */
private fun saveActionData(
    actionsInSequenceBar: List<BaseAction>,
    selectedActionIndex: Int,
    icon: ImageVector? = null,
): (String) -> Unit {
    return { name ->
        val selectedAction = actionsInSequenceBar[selectedActionIndex]

        // If icon is provided, update it, otherwise leave it as is
        icon?.let { selectedAction.icon.value = icon }

        selectedAction.data.value = name
        selectedAction.cardColor.value = specialActionColor
        selectedAction.isActive.value = true
    }
}

/**
 * Shortens the given text to a specified maximum length, adding a suffix if truncation occurs.
 *
 * @param text The original text to be minimized.
 * @param maxLength The maximum allowed length for the text, including the suffix.
 * @param suffix The string to append to the truncated text (default is "...").
 * @return The minimized text, shortened if necessary, with the suffix added if truncation occurs.
 */
fun minimizeText(text: String, maxLength: Int, suffix: String = "..."): String {
    return if (text.length > maxLength) text.take(maxLength - suffix.length) + suffix else text
}


private fun Modifier.invertedScroll(scrollState: ScrollState, enabled: Boolean = true): Modifier {
    return this.then(
        Modifier.horizontalScroll(
            scrollState,
            reverseScrolling = true,
            enabled = enabled
        )
    )
}

private fun sequenceBarTotalDuration(actionsInSequenceBar: List<BaseAction>): Float {
    val sum = actionsInSequenceBar.sumOf {it.duration.toDouble()}
    return sum.toFloat()
}

@Composable
fun DurationCountdown(
    duration: Float,
    modifier: Modifier = Modifier
) {
    Text(
        text = String.format("%.1fs", duration),
        color = Color.White,
        fontWeight = FontWeight.Bold,
        modifier = modifier
    )
}


/**
 * Sorts the sequence to ensure that loop ends don't occur before their corresponding loop begin.
 *
 * @param actionsInSequenceBar A mutable list (state) of actions to sort.
 */
private fun sortLoopButtons(actionsInSequenceBar: SnapshotStateList<BaseAction>) {
    var loopBeginIndex = 0 // Current left index
    var loopEndIndex = actionsInSequenceBar.size - 1 // Current right index
    var counterBegin = 0 // Keeps track of loop starts encountered
    var counterEnd = 0 // Keeps track of loop ends encountered

    // Keep 2 indices, iterate from both ends until they reach eachother
    while (loopBeginIndex < loopEndIndex) {
        if (actionsInSequenceBar[loopBeginIndex].name == "LOOP END" && actionsInSequenceBar[loopEndIndex].name == "LOOP START") {
            if (counterBegin == 0 && counterEnd == 0) {
                // Encountered a loop begin and corresponding end in incorrect order, swap them
                // TODO: Check why add does not do the expected effect
                val temp = actionsInSequenceBar[loopBeginIndex]
                actionsInSequenceBar[loopBeginIndex] = actionsInSequenceBar[loopEndIndex]
                actionsInSequenceBar[loopEndIndex] = temp
            }

            if (counterBegin != 0) {
                loopBeginIndex++
                counterBegin--
            }

            if (counterEnd != 0) {
                loopEndIndex--
                counterEnd--
            }
        }

        // If the left index has encountered a loop end, stay there, otherwise keep going
        if (actionsInSequenceBar[loopBeginIndex].name != "LOOP END") {
            // If the left index is on a loop begin, increment  the loop begin counter
            if (actionsInSequenceBar[loopBeginIndex].name == "LOOP START")
                counterBegin++

            loopBeginIndex++
        }

        // If the right index has encountered a loop start, stay there, otherwise keep going
        if (actionsInSequenceBar[loopEndIndex].name != "LOOP START") {
            // If the right index is on a loop end, increment  the loop begin counter
            if (actionsInSequenceBar[loopEndIndex].name == "LOOP END")
                counterEnd++

            loopEndIndex--
        }
    }
}
